include("../../AoC2021.jl")

##

hexmap = Dict(
    '0' => BitVector([0, 0, 0, 0]),
    '1' => BitVector([0, 0, 0, 1]),
    '2' => BitVector([0, 0, 1, 0]),
    '3' => BitVector([0, 0, 1, 1]),
    '4' => BitVector([0, 1, 0, 0]),
    '5' => BitVector([0, 1, 0, 1]),
    '6' => BitVector([0, 1, 1, 0]),
    '7' => BitVector([0, 1, 1, 1]),
    '8' => BitVector([1, 0, 0, 0]),
    '9' => BitVector([1, 0, 0, 1]),
    'A' => BitVector([1, 0, 1, 0]),
    'B' => BitVector([1, 0, 1, 1]),
    'C' => BitVector([1, 1, 0, 0]),
    'D' => BitVector([1, 1, 0, 1]),
    'E' => BitVector([1, 1, 1, 0]),
    'F' => BitVector([1, 1, 1, 1])
)

function parsehex(string::String)
    (collect(string) .|>
     x -> hexmap[x]) |>
    (x -> vcat(x...))
end

##

abstract type Packet end

struct LiteralValue <: Packet
    version::Int
    value::Int
end

function LiteralValue(data::BitVector)
    version = splice!(data, 1:3)' * 2 .^ collect(2:-1:0)
    splice!(data, 1:3)

    value = 0

    section = splice!(data, 1:5)
    while true
        value <<= 4
        value += section[2:end]' * 2 .^ collect(3:-1:0)

        if section[1] == 0
            break
        else
            section = splice!(data, 1:5)
        end
    end

    (LiteralValue(version, value), data)
end

struct Operator <: Packet
    version::Int
    type::Int
    subpackets::Vector{Packet}
end

function Operator(data::BitVector)
    version = splice!(data, 1:3)' * 2 .^ collect(2:-1:0)

    type = splice!(data, 1:3)' * 2 .^ collect(2:-1:0)

    subtype = splice!(data, 1)

    if subtype == 0
        subpacketdatalength = splice!(data, 1:15)' * 2 .^ collect(14:-1:0)
        subpacketdata = splice!(data, 1:subpacketdatalength)

        subpackets = []
        while length(subpacketdata) != 0
            (subpacket, subpacketdata) = parsepackets(subpacketdata)
            push!(subpackets, subpacket)
        end
    else
        subpacketcount = splice!(data, 1:11)' * 2 .^ collect(10:-1:0)

        subpackets = []
        for p ∈ 1:subpacketcount
            (subpacket, data) = parsepackets(data)

            push!(subpackets, subpacket)
        end
    end

    (Operator(version, type, subpackets), data)
end

evaluate(literal::LiteralValue) = literal.value

function evaluate(operator::Operator)
    if operator.type == 0
        sum(evaluate.(operator.subpackets))
    elseif operator.type == 1
        prod(evaluate.(operator.subpackets))
    elseif operator.type == 2
        minimum(evaluate.(operator.subpackets))
    elseif operator.type == 3
        maximum(evaluate.(operator.subpackets))
    elseif operator.type == 5
        evaluate(operator.subpackets[1]) > evaluate(operator.subpackets[2])
    elseif operator.type == 6
        evaluate(operator.subpackets[1]) < evaluate(operator.subpackets[2])
    elseif operator.type == 7
        evaluate(operator.subpackets[1]) == evaluate(operator.subpackets[2])
    end
end

##

function parsepackets(data::BitVector)
    type = data[4:6]' * 2 .^ collect(2:-1:0)

    if type == 4
        LiteralValue(data)
    else
        Operator(data)
    end
end

sumversions(literal::LiteralValue) = literal.version

sumversions(operator::Operator) = operator.version + sum(sumversions.(operator.subpackets))

##

function task0(file)
    data = parsehex(readline(file))

    (packet, _) = parsepackets(data)

    sumversions(packet)
end

##

verify(16, 1, 0, 16)
verify(16, 1, 1, 12)
verify(16, 1, 2, 23)
verify(16, 1, 3, 31)
verify(16, 1, 4, 877)

##

parsehex("D2FE28") |>
LiteralValue

parsehex("38006F45291200") |>
Operator

parsehex("EE00D40C823060") |>
Operator

##

function task1(file)
    data = parsehex(readline(file))

    (packet, _) = parsepackets(data)

    evaluate(packet)
end

##

verify(16, 2, 5, 3)
verify(16, 2, 6, 54)
verify(16, 2, 7, 7)
verify(16, 2, 8, 9)
verify(16, 2, 9, 1)
verify(16, 2, 10, 0)
verify(16, 2, 11, 0)
verify(16, 2, 12, 1)
verify(16, 2, 4, 194435634456)